from rest_framework.parsers import BaseParser

from rdflib import Graph


class RDFLibParser(BaseParser):
    """ Base class for RDF parsers.

    Derive from this class to create a parser for a particular RDF
    serialization format such as JSON-LD. Subclasses need not define
    any methods, but should define two static attributes.

    The first attribute is facing the Django REST Framework side of
    the equation: `media_type` (the MIME type). This is used in
    content negotiation.

    The other attribute is `rdflib_args`, which must be a dictionary
    and which should list all named arguments to Graph.parse. At the
    very least, this should include the `format` parameter, which
    determines the serialization format.
    """
    def parse(self, stream, media_type=None, parser_context=None):
        graph = Graph()
        graph.parse(data=stream.read(), **self.rdflib_args)
        return graph


class JSONLDParser(RDFLibParser):
    media_type = 'application/ld+json'
    rdflib_args = {
        'format': 'json-ld',
    }



from .utils import invalid_xml_remove

INVALID_STR = '''La trilogie a en eet
comptabilisé plus de 26 millions de vente.'''

VALID_STR = '''La trilogie a en e et
comptabilisé plus de 26 millions de vente.'''


def test_xml_sanitize():
    assert VALID_STR == invalid_xml_remove(VALID_STR)
    assert INVALID_STR[18] == chr(0x1b) == '\x1b'
    sanitized_str = invalid_xml_remove(INVALID_STR)
    assert sanitized_str[18] == chr(0x20) == ' '



from rdflib import Namespace

SOURCES_NS = Namespace('{}{}#'.format('http://testserver/', 'source/'))

NLP_ONTOLOGY_NS =  '{}{}#'.format('http://testserver/', 'nlp-ontology')


nlp = Namespace(NLP_ONTOLOGY_NS)



""" External namespaces that we use internally. """

from rdflib.namespace import Namespace, RDF, RDFS, OWL, XSD, FOAF, SKOS, DC, DCTERMS

FRBR    = Namespace('http://purl.org/vocab/frbr/core#')
FRBROO = Namespace('http://iflastandards.info/ns/fr/frbr/frbroo/')
ERLANGEN = Namespace('http://erlangen-crm.org/current/')
CIDOC   = Namespace('http://www.cidoc-crm.org/cidoc-crm/')
OA      = Namespace('http://www.w3.org/ns/oa#')
AS      = Namespace('http://www.w3.org/ns/activitystreams#')
DCTYPES = Namespace('http://purl.org/dc/dcmitype/')
SCHEMA  = Namespace('http://schema.org/')
HTTP    = Namespace('http://www.w3.org/2011/http#')
HTTPSC  = Namespace('https://www.w3.org/2011/http-statusCodes#')
HTTPM   = Namespace('https://www.w3.org/2011/http-methods#')
ISO6391 = Namespace('http://id.loc.gov/vocabulary/iso639-1/')
OWL     = Namespace('http://www.w3.org/2002/07/owl#')

# URI for representing an unauthenticated visitor, an unknown/multiple languages, or unknown source type.
# Probably not the best possible, but it will do for now.
UNKNOWN = 'https://www.wikidata.org/wiki/Q24238356'



from types import SimpleNamespace

import pytest
from django.conf import settings
from rdflib.plugins.stores.sparqlstore import SPARQLUpdateStore
from rdf.ns import RDF, SCHEMA
from rdf.utils import graph_from_triples
from rdflib import Literal
from rdf.conftest import sparqlstore
from sparql.test_app.graphs import sources_graph as graph
from sparql.test_app.constants import *


INSERT_QUERY = '''
    PREFIX my: <http://testserver/nlp-ontology#>
    PREFIX ns3: <http://schema.org/>

    INSERT DATA {
        my:icecream         a               ns3:Food        ;
                            ns3:color       "#f9e5bc"       .
        ns3:Cat             my:meow         "loud"          .
    }
'''

DELETE_DATA_QUERY = '''
    PREFIX my: <http://testserver/nlp-ontology#>
    PREFIX ns3: <http://schema.org/>

    DELETE DATA {
        my:icecream         a               ns3:Food        ;
                            ns3:color       "#f9e5bc"       .
        ns3:Cat             my:meow         "loud"          .
    }
'''

SELECT_QUERY = '''
    SELECT ?s ?p ?o
    WHERE {
        ?s ?p ?o .
    }
'''

SELECT_FROM_NLP_QUERY = '''
    SELECT ?s ?p ?o
    FROM <{ns}>
    WHERE {{
        ?s ?p ?o .
    }}
'''.format(ns=NLP_ONTOLOGY_NS)

SELECT_FROM_SOURCES_QUERY = '''
    SELECT ?s ?p ?o
    FROM <{ns}>
    WHERE {{
        ?s ?p ?o .
    }}
'''.format(ns=SOURCES_NS)

ASK_QUERY = '''PREFIX my: <http://testserver/nlp-ontology#>
ASK { ?x my:meow  "loud" }'''
ASK_QUERY_FALSE = '''PREFIX my: <http://testserver/nlp-ontology#>
ASK { ?x my:meow  "silent" }'''

CONSTRUCT_QUERY = '''
    PREFIX my: <http://testserver/nlp-ontology#>
    PREFIX ns3: <http://schema.org/>
    CONSTRUCT WHERE { ?x my:meow ?name }
'''

DELETE_QUERY = '''
    PREFIX my: <{ns}>
    DELETE {{ ?s ?p ?o }} WHERE {{ ?s ?p ?o ; my:meow "loud" }}
'''.format(ns=NLP_ONTOLOGY_NS)

DELETE_FROM_QUERY = '''
    PREFIX my: <{ns}>
    DELETE {{ GRAPH <{graph}> {{ ?s ?p ?o }} }} WHERE {{ ?s ?p ?o ; my:meow "loud" }}
'''.format(ns=NLP_ONTOLOGY_NS, graph=NLP_ONTOLOGY_NS)


@pytest.fixture
def triples():
    return (
        (nlp.icecream,   RDF.type,       SCHEMA.Food),
        (nlp.icecream,   SCHEMA.color,   Literal("#f9e5bc")),
        (SCHEMA.Cat,    nlp.meow,        Literal("loud")),
    )


@pytest.fixture
def ontologygraph(triples):
    g = graph_from_triples(triples)
    return g


@pytest.fixture
def graph_db(db, sparqlstore):
    return graph()


@pytest.fixture
def ontologygraph_db(db, ontologygraph, sparqlstore):
    g = graph()
    g += ontologygraph
    yield
    g -= ontologygraph


@pytest.fixture
def test_queries():
    values = {
        'ASK_TRUE': ASK_QUERY,
        'ASK_FALSE': ASK_QUERY_FALSE,
        'CONSTRUCT': CONSTRUCT_QUERY,
        'SELECT': SELECT_QUERY,
        'SELECT_FROM_NLP': SELECT_FROM_NLP_QUERY,
        'SELECT_FROM_SOURCES': SELECT_FROM_SOURCES_QUERY,
        'INSERT': INSERT_QUERY,
        'DELETE': DELETE_QUERY,
        'DELETE_DATA': DELETE_DATA_QUERY,
        'DELETE_FROM': DELETE_FROM_QUERY
    }
    return SimpleNamespace(**values)


@ pytest.fixture
def accept_headers():
    values = {
        'turtle': 'text/turtle',
        'sparql_json': 'application/sparql-results+json',
        'sparql_xml': 'application/sparql-results+xml',
        'sparql_csv': 'text/csv',
        'json': 'application/json',
        'rdfxml': 'application/rdf+xml',
        'ntriples': 'application/n-triples',
        'jsonld': 'application/ld+json'
    }
    return SimpleNamespace(**values)

@pytest.fixture
def unsupported_queries():
    queries = {
        'LOAD': 'LOAD <{}> INTO GRAPH <{}>'.format(SOURCES_NS, NLP_ONTOLOGY_NS),
        'CLEAR_ALL': 'CLEAR ALL',
        'CLEAR': 'CLEAR GRAPH <{}>'.format(NLP_ONTOLOGY_NS),
        'DROP': 'CLEAR GRAPH <{}>'.format(NLP_ONTOLOGY_NS),
        'ADD': 'ADD <{}> TO <{}>'.format(SOURCES_NS, NLP_ONTOLOGY_NS),
        'MOVE': 'MOVE <{}> TO <{}>'.format(SOURCES_NS, NLP_ONTOLOGY_NS),
        'COPY': 'COPY <{}> TO <{}>'.format(SOURCES_NS, NLP_ONTOLOGY_NS),
        'CREATE': 'CREATE GRAPH <http://testserver/newgraph#>'
    }

    return queries


@pytest.fixture
def blanknode_queries():
    triples = (
        ('[ :p "v" ]'),
        ('[] :p "v"'),
        ('_:b57 :p "v"'),
    )
    return ['INSERT DATA {{ {} }}'.format(t) for t in triples]




from rdflib import URIRef, Graph

from .migrations import *


class TestMigration(RDFMigration):
    def actual(self):
        return Graph()

    def desired(self):
        return Graph()

    @on_add('http://example.com/')
    def process_addition(self, actual, conjunctive):
        pass

    @on_remove('http://other-example.com/')
    def process_deletion(self, actual, conjunctive):
        pass


def test_RDFMigration():
    tester = TestMigration()
    assert len(tester.actual()) == 0
    assert len(tester.desired()) == 0
    assert tester.add_handlers == {
        URIRef('http://example.com/'): 'process_addition',
    }
    assert tester.remove_handlers == {
        URIRef('http://other-example.com/'): 'process_deletion',
    }



import re

from django.conf import settings

# These update operations are not supported
# TODO: wiki reference
UPDATE_NOT_SUPPORTED = ['Load', 'Clear',
                        'Drop', 'Add', 'Move', 'Copy', 'Create']
UPDATE_NOT_SUPPORTED_PATTERN = re.compile(
    '|'.join(UPDATE_NOT_SUPPORTED), re.IGNORECASE)

BLANK_NODE_PATTERN = re.compile(r'\[.*\]|_:')

SPARQL_NS = '{}/sparql'.format(settings.RDF_NAMESPACE_ROOT)



from io import BytesIO

from rdflib import Graph

from .parsers import JSONLDParser
from .ns import *


def test_rdflibparser(filled_graph):
    parser = JSONLDParser()
    serialized = filled_graph.serialize(format='json-ld').encode()
    parsed = parser.parse(BytesIO(serialized))
    assert len(parsed ^ filled_graph) == 0



from django.db import models
from django.db.models import F
from django.db.transaction import atomic
# See https://docs.djangoproject.com/en/3.2/_modules/django/utils/functional
from django.utils.functional import classproperty


class BaseCounter(models.Model):
    class Meta:
        abstract = True

    """ AUTOINCREMENT for RDF subject URIs. """
    count = models.PositiveIntegerField()

    @property
    def namespace(self):
        raise NotImplementedError()

    @classproperty
    def current(cls):
        """ Get or create a singleton instance of ItemCounter. """
        instance = cls.objects.all().first()
        if not instance:
            instance = cls(count=1)
            instance.save()
        return instance

    def __str__(self):
        """ The subject URI associated with the current value of `count`. """
        return '{}{}'.format(self.namespace, self.count)

    @atomic
    def increment(self):
        """ Add 1 to the count and save immediately. """
        self.count = F('count') + 1
        self.save()
        self.refresh_from_db()



from rdf.renderers import (JsonLdRenderer, NTriplesRenderer, RdfXMLRenderer,
                           TurtleRenderer)
from rest_framework.negotiation import DefaultContentNegotiation

from .renderers import (QueryResultsCSVRenderer, QueryResultsJSONRenderer,
                        QueryResultsXMLRenderer)


class SPARQLContentNegotiator(DefaultContentNegotiation):
    results_renderers = [QueryResultsJSONRenderer, QueryResultsXMLRenderer,
                         QueryResultsCSVRenderer]
    rdf_renderers = [TurtleRenderer, RdfXMLRenderer,
                     JsonLdRenderer, NTriplesRenderer]
    querytype_accepts = {
        'SELECT': results_renderers,
        'ASK': results_renderers,
        'CONSTRUCT': rdf_renderers,
        'EMPTY': rdf_renderers
    }

    def select_renderer(self, request, renderers, format_suffix=None):
        query_type = request.data.get('query_type', None)
        renderer_classes = self.querytype_accepts.get(
            query_type, self.rdf_renderers + self.results_renderers)
        renderers = [renderer() for renderer in renderer_classes]

        return super().select_renderer(request, renderers, format_suffix)



import json

import pytest
from rdf.ns import SCHEMA
from rdflib.namespace import Namespace
from rdf.utils import graph_from_triples
from rdflib import XSD, Graph, Literal

from .exceptions import BlankNodeError
from .views import SPARQLUpdateAPIView
from .test_app.views import QueryView, UpdateView, NLPQueryView
from .conftest import nlp

from rest_framework.test import APIRequestFactory

QUERY_URL = '/test/query'
UPDATE_URL = '/test/update'

class MockClient:
    '''
    Class used to test a view. Can be used like the 
    api test client, but calls the view directly,
    without routing or permissions.
    '''

    request_factory = APIRequestFactory()

    def __init__(self, view_class):
        self.view = view_class.as_view()

    def get(self, path, data = None, **kwargs):
        request = self.request_factory.get(path, data, **kwargs)
        return self.view(request).render()

    def post(self, path, data = None, **kwargs):
        request = self.request_factory.post(path, data, **kwargs)
        return self.view(request).render()

@pytest.fixture
def query_client():
    return MockClient(QueryView)

@pytest.fixture
def update_client():
    return MockClient(UpdateView)

@pytest.fixture
def nlp_query_client():
    return MockClient(NLPQueryView)

def check_content_type(response, content_type):
    return content_type in response.headers['content-type']

def test_insert(query_client, update_client, ontologygraph, test_queries, graph_db):
    assert len(graph_db) == 0

    post_response = update_client.post(
        UPDATE_URL, {'update': test_queries.INSERT})
    assert post_response.status_code == 200

    get_response = query_client.get(QUERY_URL)
    assert get_response.status_code == 200
    assert check_content_type(get_response, 'text/turtle')

    get_data = Graph().parse(data=get_response.content, format='text/turtle')
    assert len(graph_db) != 0
    assert len(get_data ^ ontologygraph) == 0

    # clean up
    update_client.post(
        UPDATE_URL, {'update': test_queries.DELETE_DATA})
    assert len(graph_db) == 0


def test_ask(query_client, test_queries, ontologygraph_db):
    true_response = query_client.get(QUERY_URL, {'query': test_queries.ASK_TRUE})
    assert true_response.status_code == 200
    assert json.loads(true_response.content.decode('utf8'))['boolean']
    assert check_content_type(true_response, 'application/sparql-results+json')

    false_response = query_client.get(
        QUERY_URL, {'query': test_queries.ASK_FALSE})
    assert false_response.status_code == 200
    assert not json.loads(false_response.content.decode('utf8'))['boolean']


def test_construct(query_client, test_queries, ontologygraph_db):
    response = query_client.get(QUERY_URL, {'query': test_queries.CONSTRUCT})
    assert response.status_code == 200

    result_graph = Graph().parse(data=response.content, format='turtle')
    exp_graph = graph_from_triples(
        ((SCHEMA.Cat,    nlp.meow,        Literal("loud", datatype=XSD.string)),)
    )
    assert len(exp_graph ^ result_graph) == 0


def test_malformed(query_client, update_client, sparqlstore):
    malformed_get = query_client.post(
        QUERY_URL, {'query': 'this is no SPARQL query!'})
    assert malformed_get.status_code == 400

    malformed_update = update_client.post(
        UPDATE_URL, {'update': 'this is no SPARQL query!'})
    assert malformed_update.status_code == 400


def test_unsupported(update_client, unsupported_queries, sparqlstore):
    for query in unsupported_queries.values():
        request = update_client.post(UPDATE_URL, {'update': query})
        assert b'Update operation is not supported.' in request.content
        assert request.status_code == 400


def test_delete(query_client, update_client, test_queries, ontologygraph_db, graph_db):
    # Should not delete if from another endpoint
    delete = update_client.post(
        '/sparql/source/update', {'update': test_queries.DELETE_FROM})
    assert delete.status_code == 400
    res = query_client.get(QUERY_URL).content
    assert len(Graph().parse(data=res, format='turtle')) == 3

    # Should delete if endpoint and graph match
    assert len(graph_db) != 2
    delete = update_client.post(
        UPDATE_URL, {'update': test_queries.DELETE})
    assert delete.status_code == 200
    res = query_client.get(QUERY_URL).content
    assert len(Graph().parse(data=res, format='turtle')) == 2
    assert len(graph_db) == 2


def test_select_from(query_client, nlp_query_client, test_queries, ontologygraph_db, ontologygraph, accept_headers):
    # Should not return results if querying a different endpoint
    # Note that any triples in VOCAB_NS graph would be returned here
    res = nlp_query_client.post('/sparql/vocab/query',
                             {'query': test_queries.SELECT_FROM_NLP})
    assert res.status_code == 200
    assert len(res.data) == 0

    # Should ignore FROM clause if endpoint and graph don't match
    res = query_client.post(QUERY_URL,
                             {'query': test_queries.SELECT_FROM_SOURCES})
    assert res.status_code == 200
    assert len(res.data) == 3

    # Should return results if FROM and endpoint match
    res = query_client.post(QUERY_URL,
                             {'query': test_queries.SELECT_FROM_NLP})
    assert res.status_code == 200
    assert len(res.data) == 3


def test_blanknodes(blanknode_queries):
    view = SPARQLUpdateAPIView()
    for q in blanknode_queries:
        with pytest.raises(BlankNodeError):
            view.check_supported(q)


def test_GET_body(query_client):
    res = query_client.get(
        QUERY_URL,
        data={'query': 'some query'}
    )
    assert res.status_code == 400


def test_graph_negotiation(query_client, accept_headers, test_queries):
    ''' result headers should succeed for CONSTRUCT/DESRIBE/empty, fail for SELECT/ASK'''
    graph_headers = [accept_headers.rdfxml, 
                     accept_headers.jsonld, accept_headers.turtle]

    for h in graph_headers:
        accept = query_client.get(
            QUERY_URL, {'query': test_queries.CONSTRUCT}, HTTP_ACCEPT=h)
        deny = query_client.get(
            QUERY_URL, {'query': test_queries.SELECT}, HTTP_ACCEPT=h)
        empty = query_client.get(QUERY_URL, HTTP_ACCEPT=h)
        assert accept.status_code == 200
        assert check_content_type(accept, h)
        assert deny.status_code == 406
        assert empty.status_code == 200
        assert check_content_type(empty, h)


def test_result_negotiation(query_client, accept_headers, test_queries):
    ''' result headers should succeed for SELECT/ASK, fail for CONSTRUCT/DESCRIBE/empty'''
    result_headers = [accept_headers.sparql_json, accept_headers.sparql_xml,
                      accept_headers.sparql_csv]
    for h in result_headers:
        accept = query_client.get(
            QUERY_URL, {'query': test_queries.SELECT}, HTTP_ACCEPT=h)
        deny = query_client.get(
            QUERY_URL, {'query': test_queries.CONSTRUCT}, HTTP_ACCEPT=h)
        empty = query_client.get(QUERY_URL, HTTP_ACCEPT=h)
        assert accept.status_code == 200
        assert check_content_type(accept, h)
        assert deny.status_code == 406
        assert empty.status_code == 406

    # application/json should not work
    regular_json = query_client.get(
        QUERY_URL, {'query': test_queries.SELECT},
        HTTP_ACCEPT=accept_headers.json)
    assert regular_json.status_code == 406



from sparql.views import SPARQLUpdateAPIView, SPARQLQueryAPIView
from rest_framework.permissions import IsAdminUser
from .graphs import sources_graph, nlp_graph

class QueryView(SPARQLQueryAPIView):
    def graph(self):
        return sources_graph()
    
class UpdateView(SPARQLUpdateAPIView):

    def graph(self):
        return sources_graph()

class NLPQueryView(SPARQLQueryAPIView):
    def graph(self):
        return nlp_graph()


import pytest

from rdflib import Graph

from .ns import *
from .utils import *
from .conftest import ITEM
import re


@pytest.fixture
def other_triples():
    return (
        (RDF.type, RDF.type, RDF.Property),
        (RDF.type, RDFS.range, RDFS.Class),
        (RDF.type, RDFS.label, Literal('typo')),  # sic
    )


def test_prune_triples(filled_graph, other_triples):
    before = len(filled_graph)
    prune_triples(filled_graph, other_triples)
    after = len(filled_graph)
    assert before - after == 2


def test_prune_triples_cascade(filled_conjunctive_graph):
    anno = (ITEM['7'], RDF.type, OA.Annotation)
    # get our item graph from the conjunctive graph
    privileged_graph = next(
        filled_conjunctive_graph.contexts())  # victim graph :D
    prune_triples_cascade(filled_conjunctive_graph,
                          (anno,), [privileged_graph])
    assert len(filled_conjunctive_graph) == 0


def test_prune_triples_cascade_privileged(filled_conjunctive_graph):
    anno = (ITEM['7'], RDF.type, OA.Annotation)
    # get our item graph from the conjunctive graph
    privileged_graph = next(
        filled_conjunctive_graph.contexts())  # victim graph :D
    prune_triples_cascade(filled_conjunctive_graph, (anno,), [
                          privileged_graph], [OA.hasBody])
    assert len(filled_conjunctive_graph) == 14


def test_prune_zero(filled_graph):
    prune_triples(filled_graph, filled_graph)
    assert len(filled_graph) == 0


def test_append_triples(filled_graph, other_triples):
    before = len(filled_graph)
    append_triples(filled_graph, other_triples)
    after = len(filled_graph)
    assert after - before == 1


def test_append_identity(filled_graph):
    before = len(filled_graph)
    append_triples(filled_graph, filled_graph)
    after = len(filled_graph)
    assert before == after


def test_graph_from_triples(other_triples):
    result = graph_from_triples(other_triples)
    assert isinstance(result, Graph)
    for t in other_triples:
        assert t in result


def test_traverse_forward(filled_graph, other_triples):
    start = graph_from_triples(other_triples)
    result = traverse_forward(filled_graph, start, 0)
    assert isinstance(result, Graph)
    size = len(result)
    assert size == 0
    for depth in range(1, 4):
        result = traverse_forward(filled_graph, start, depth)
        new_size = len(result)
        assert new_size > size
        size = new_size


def test_traverse_backward(filled_graph, other_triples):
    start = graph_from_triples(other_triples)
    result = traverse_backward(filled_graph, start, 0)
    assert isinstance(result, Graph)
    size = len(result)
    assert size == 0
    for depth in range(1, 4):
        result = traverse_backward(filled_graph, start, depth)
        new_size = len(result)
        assert new_size > size
        size = new_size


def test_prefix_injection(sparqlstore, prefixed_query):
    expected_prefixes = ['rdf', 'rdfs', 'schema']

    res = sparqlstore._inject_prefixes(prefixed_query, {})
    assert all(prefix in re.findall(PREFIX_PATTERN, res)
               for prefix in expected_prefixes)
    assert 'PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>' in res

    res = sparqlstore._inject_prefixes(
        prefixed_query, extra_bindings={'rdf': 'https://cat-bounce.com',
                                        'schema': 'http://randomcolour.com/'})
    assert all(prefix in re.findall(PREFIX_PATTERN, res)
               for prefix in expected_prefixes)
    assert 'PREFIX rdf: <https://cat-bounce.com>' in res
    assert 'PREFIX schema: <http://randomcolour.com/>' in res


def test_charset_update(sparqlstore):
    problematic_triple = (RDF.type, RDFS.label, Literal('Février'))
    sparqlstore.add(problematic_triple)

    # If the triple is properly stored, it can be retrieved
    # otherwise it will be stored as 'FÃ©vrier'
    assert next(sparqlstore.triples(problematic_triple), None)



from rest_framework.renderers import BaseRenderer


class RDFLibRenderer(BaseRenderer):
    """ Base class for RDF renderers.

    Derive from this class to create a renderer for a particular RDF
    serialization format such as JSON-LD. Subclasses need not define
    any methods, but should define two groups of static attributes.

    The first group is facing the Django REST Framework side of the
    equation. These are the same attributes that all BaseRenderer
    subclasses should define: `media_type` (the MIME type), `format`
    (an extension that may be append to the request path as an
    alternative means of content negotiation) and optionally `charset`.
    They are used in content negotiation.

    The other group is the single `rdflib_args` attribute, which must
    be a dictionary and which should list all named arguments to
    Graph.serialize. At the very least, this should include the
    `format` parameter, which determines the serialization format.
    """

    def render(self, graph, media_type=None, renderer_context=None):
        view = None
        if renderer_context is not None and 'view' in renderer_context:
            view = renderer_context['view']
        rdflib_args = self.get_rdflib_args(view=view)
        return graph.serialize(**rdflib_args)

    def get_rdflib_args(self, **kwargs):
        return self.rdflib_args


class TurtleRenderer(RDFLibRenderer):
    media_type = 'text/turtle'
    format = 'ttl'
    rdflib_args = {
        'format': 'turtle',
    }


class RdfXMLRenderer(RDFLibRenderer):
    media_type = 'application/rdf+xml'
    format = 'xml'
    rdflib_args = {
        'format': 'xml',
    }


class JsonLdRenderer(RDFLibRenderer):
    media_type = 'application/ld+json'
    format = 'jsonld'
    json_ld_context = None

    def get_rdflib_args(self, view=None):
        args = {
            'format': 'json-ld',
        }
        if (
                hasattr(view, "json_ld_context")
                and view.json_ld_context is not None
        ):
            args['context'] = view.json_ld_context
        return args


class NTriplesRenderer(RDFLibRenderer):
    media_type = 'application/n-triples'
    format = 'nt'
    rdflib_args = {
        'format': 'nt',
        'encoding': 'ascii'  # N-triples are always ascii encoded
    }



import sys
import re
from rdflib import Literal
from rdf.ns import XSD

# via http://stackoverflow.com/questions/1707890/fast-way-to-filter-illegal-xml-unicode-chars-in-python
ILLEGAL_UNICHRS = [
    (0x00, 0x08), (0x0B, 0x1F), (0x7F, 0x84), (0x86, 0x9F),
    (0xD800, 0xDFFF), (0xFDD0, 0xFDDF), (0xFFFE, 0xFFFF),
    (0x1FFFE, 0x1FFFF), (0x2FFFE, 0x2FFFF), (0x3FFFE, 0x3FFFF),
    (0x4FFFE, 0x4FFFF), (0x5FFFE, 0x5FFFF), (0x6FFFE, 0x6FFFF),
    (0x7FFFE, 0x7FFFF), (0x8FFFE, 0x8FFFF), (0x9FFFE, 0x9FFFF),
    (0xAFFFE, 0xAFFFF), (0xBFFFE, 0xBFFFF), (0xCFFFE, 0xCFFFF),
    (0xDFFFE, 0xDFFFF), (0xEFFFE, 0xEFFFF), (0xFFFFE, 0xFFFFF),
    (0x10FFFE, 0x10FFFF)
]
ILLEGAL_RANGES = ["%s-%s" % (chr(low), chr(high))
                  for (low, high) in ILLEGAL_UNICHRS if low < sys.maxunicode]

ILLEGAL_XML_RE = re.compile(u'[%s]' % u''.join(ILLEGAL_RANGES))

CLEANABLE_DATATYPES = (None, XSD.string, XSD.normalizedString)


def invalid_xml_remove(c):
    return re.sub(ILLEGAL_XML_RE, ' ', c)


def find_invalid_xml(c):
    return re.search(ILLEGAL_XML_RE, c)


def is_cleanable(term):
    return isinstance(term, Literal) and term.datatype in CLEANABLE_DATATYPES


def clean_term(term):
    if not is_cleanable(term):
        return term
    return Literal(invalid_xml_remove(term.value),
                   lang=term.language,
                   datatype=term.datatype)


def xml_sanitize_triple(triple):
    if any((is_cleanable(term)) for term in triple):
        cleaned_triple = tuple([clean_term(t) for t in triple])
        if cleaned_triple != triple:
            return (True, cleaned_triple)
    return (False, triple)



import json

from rdflib import Graph
from rest_framework.renderers import BaseRenderer

from .renderers import TurtleRenderer, JsonLdRenderer
from .ns import *
from .views import RDFView


def verify_serialization(serialization, graph: Graph, format: str) -> bool:
    parsed = Graph()
    parsed.parse(data=serialization, format=format)
    return len(parsed ^ graph) == 0


def test_rdflibrenderer(filled_graph):
    renderer = TurtleRenderer()
    serialization = renderer.render(filled_graph)
    assert verify_serialization(serialization, filled_graph, "turtle")


def test_jsonldrenderer_no_context_arg(filled_graph):
    # Using the JsonLdRenderer with no context argument at all should work
    renderer = JsonLdRenderer()
    serialization = renderer.render(filled_graph)
    assert verify_serialization(serialization, filled_graph, "json-ld")


def test_jsonldrenderer_no_jsonld_context(filled_graph):
    view = RDFView()
    view.json_ld_context = None
    renderer = JsonLdRenderer()
    context = {"view": view}
    serialization = renderer.render(filled_graph, renderer_context=context)
    assert verify_serialization(serialization, filled_graph, "json-ld")


def test_jsonldrenderer_with_jsonld_context(filled_graph):
    # Note that this test does not check if the JSON-LD context is correctly
    # applied - we assume that rdflib does this correctly.
    view = RDFView()
    view.json_ld_context = {
        "schema": "https://schema.org/",
    }
    renderer = JsonLdRenderer()
    context = {"view": view}
    serialization = renderer.render(filled_graph, renderer_context=context)
    ser_dict = json.loads(serialization)
    assert isinstance(ser_dict, dict)
    assert "@context" in ser_dict
    assert verify_serialization(serialization, filled_graph, "json-ld")




"""
Unittests for the rdfmigrate command.

In the test code, you find multiple expressions of the following form:

    len(left ^ right) == 0

where `left` and `right` are graphs. In short, this expression
evaluates to true if and only if both graphs contain the exact same
set of triples. Please beware that this is not a completely valid way
of comparison when blank nodes are involved!
"""

import pytest

from importlib import import_module
from sys import stdout

from rdflib import Graph

from rdf.utils import graph_from_triples
from rdf.ns import RDF, RDFS
from rdf.migrations import *
from .rdfmigrate import *


@pytest.mark.django_db
def test_migrate_package(app_with_rdf_migrations, app_without_rdf_migrations):
    command = Command()
    command.migrate_package(app_with_rdf_migrations)
    fixture = import_module('.fixture', app_with_rdf_migrations)
    graph = import_module('.graph', app_with_rdf_migrations)
    assert len(graph.graph() ^ fixture.canonical_graph()) == 0
    command.migrate_package(app_without_rdf_migrations)
    # We expect the above line to be a no-op.
    # It must not throw an exception.
    assert True


@pytest.mark.django_db
def test_migrate_graph(empty_graph, filled_graph):
    command = Command()
    backup = graph_from_triples(filled_graph)
    extra = (RDFS.label, RDFS.domain, RDFS.Resource)

    class TestMigration(RDFMigration):
        def actual(self):
            return empty_graph
        def desired(self):
            return filled_graph
        @on_add(RDF.type)
        def insert_extra(self, actual, conjunctive):
            actual.add(extra)

    command.migrate_graph(TestMigration())
    assert len(filled_graph ^ backup) == 0
    assert len(backup - empty_graph) == 0
    assert len(empty_graph - backup) == 1
    assert extra in empty_graph



import json
from io import StringIO

from rdf.ns import HTTP, HTTPSC, RDF
from rdf.renderers import RDFLibRenderer, TurtleRenderer
from rdf.utils import graph_from_triples
from rdflib import BNode, Literal
from rdflib.plugins.sparql.results.jsonresults import JSONResultSerializer
from rest_framework.renderers import JSONRenderer


class QueryResultsTurtleRenderer(TurtleRenderer):
    ''' Renders turtle from rdflib SPARQL query results'''

    def render(self, query_results, media_type=None, renderer_context=None):
        results_graph = graph_from_triples(query_results)
        return super().render(results_graph, media_type, renderer_context)


class QueryResultsJSONRenderer(JSONRenderer):
    ''' Renders SPARQL Query Results JSON Format from rdflib query results'''
    media_type = 'application/sparql-results+json'
    format = 'srj'

    def render(self, query_results, media_type=None, renderer_context=None):
        with StringIO() as stream:
            JSONResultSerializer(query_results).serialize(stream)
            json_str = stream.getvalue()
            serialized_results = json.loads(json_str)
            return super().render(serialized_results,
                                  media_type,
                                  renderer_context)


class QueryResultsXMLRenderer(RDFLibRenderer):
    media_type = 'application/sparql-results+xml'
    format = 'xml'

    def render(self, query_results, media_type=None, renderer_context=None):
        return query_results.serialize(format='xml')


class QueryResultsCSVRenderer(RDFLibRenderer):
    media_type = 'text/csv'
    format = 'csv'
    rdflib_args = {
        'format': 'csv'
    }



from rest_framework.exceptions import APIException, ParseError


class NoParamError(APIException):
    status_code = 400
    default_detail = 'No SPARQL-Query or SPARQL-Update in query or update body parameters.'
    default_code = 'sparql_no_param_error'


class ParseSPARQLError(ParseError):
    default_detail = 'Error parsing SPARQL.'
    default_code = 'sparql_parse_error'


class UnsupportedUpdateError(ParseError):
    default_detail = 'SPARQL-Update operation not supported.'
    default_code = 'sparql_unsupported_update'


class BlankNodeError(ParseError):
    default_detail = 'Blank nodes are not supported.'
    default_code = 'sparql_blanknode'



import random
import re
from urllib.parse import urlencode
from urllib.request import Request, urlopen

from django.conf import settings
from rdflib import ConjunctiveGraph, Graph, Literal, URIRef
from rdflib.plugins.stores.sparqlstore import SPARQLStore
from rdflib.plugins.stores.sparqlconnector import (
    SPARQLConnector, SPARQLConnectorException, _response_mime_types)
from typing import Optional
from .ns import OA, XSD, DCTERMS


PREFIX_PATTERN = re.compile(r'PREFIX\s+(\w+):\s*<\S+>', re.IGNORECASE)


def get_conjunctive_graph():
    """ Returns the conjunctive graph of our SPARQL store. """
    return ConjunctiveGraph(settings.RDFLIB_STORE)


def prune_triples(graph, triples):
    """Remove all items in iterable `triples` from `graph` (modify in place)."""
    for triple in triples:
        graph.remove(triple)


def prune_triples_cascade(graph, triples, graphs_applied_to=[], privileged_predicates=[]):
    """
    Recursively remove subjects in `triples` and all related resources from `graph`.
    Specify which graphs qualify, i.e. from which triples will be deleted, in `graphs_applied_to`.
    Optionally, skip items related via specific (privileged) predicates.
    """
    for triple in triples:
        prune_recursively(
            graph, triple[0], graphs_applied_to, privileged_predicates
        )


def prune_recursively(graph, subject, graphs_applied_to=[], privileged_predicates=[]):
    """
    Recursively remove subject and all related resources from `graph`.
    Specify which graphs qualify, i.e. from which triples will be deleted, in `graphs_applied_to`.
    Optionally, skip deletion of (i.e. keep) items related via specific (privileged) predicates.
    """
    related_by_subject = list(graph.quads((subject, None, None)))

    for s, p, o, c in related_by_subject:
        if isinstance(o, URIRef) and o != s and p not in privileged_predicates and c in graphs_applied_to:
            prune_recursively(graph, o, graphs_applied_to,
                              privileged_predicates)

    prune_triples(graph, related_by_subject)


def append_triples(graph, triples):
    """ Add all items in iterable `triples` to `graph` (modify in place). """
    for triple in triples:
        graph.add(triple)


def graph_from_triples(triples, ctor=Graph):
    """ Return a new Graph containing all items in iterable `triples`. """
    graph = ctor()
    append_triples(graph, triples)
    return graph


def sample_graph(graph, subjects, request):
    """ Return a random sample from a graph, optionally filtering with a list containing [predicate, object]. """
    n_results = int(request.GET.get('n_results'))
    if len(subjects) > n_results:
        sampled_subjects = random.sample(list(subjects), n_results)
    else:
        sampled_subjects = subjects
    output = Graph()
    for sub in sampled_subjects:
        suggestions = graph.triples((sub, None, None))
        [output.add(s) for s in suggestions]
    return output


def traverse_forward(full_graph, fringe, plys):
    """
    Traverse `full_graph` by object `plys` times, starting from `fringe`.

    Returns a graph with all triples accumulated during the traversal,
    excluding `fringe`.
    """
    result = Graph()
    visited_objects = set()
    while plys > 0:
        objects = set(fringe.objects()) - visited_objects
        if not len(objects):
            break
        fringe = Graph()
        for o in objects:
            if not isinstance(o, Literal):
                append_triples(fringe, full_graph.triples((o, None, None)))
        result |= fringe
        visited_objects |= objects
        plys -= 1
    return result


def traverse_backward(full_graph, fringe, plys):
    """
    Traverse `full_graph` by subject `plys` times, starting from `fringe`.

    Returns a graph with all triples accumulated during the traversal,
    excluding `fringe`. This result always contains complete
    resources, i.e., all triples of each subject in the graph are
    included.
    """
    result = Graph()
    subjects = set(fringe.subjects())
    visited_subjects = set()
    while plys > 0:
        if not len(subjects):
            break
        fringe = Graph()
        fringe_subjects = set()
        for s in subjects:
            parents = set(full_graph.subjects(None, s))
            for ss in parents - fringe_subjects:
                append_triples(fringe, full_graph.triples((ss, None, None)))
            fringe_subjects |= parents
        result |= fringe
        visited_subjects |= subjects
        subjects = set(fringe.subjects()) - visited_subjects
        plys -= 1
    return result


def latin1_to_utf8(original: str) -> str:
    try:
        return original.encode('latin-1').decode()
    except (UnicodeDecodeError, UnicodeEncodeError):
        return original


def find_latin1_triples(graph: Graph) -> Graph:
    query = r'''CONSTRUCT {{?s ?p ?o }}
    WHERE {{
        GRAPH <{}> {{
            ?s ?p ?o;
            dcterms:created ?date .
            FILTER(?date > "2022-05-10T00:00:00.000000+00:00"^^xsd:dateTime)
            FILTER(datatype(?o)=xsd:string)
            FILTER(regex(?o, "[\\x80-\\xFF]"))
        }}
    }}
    '''.format(graph.identifier)
    res = graph.query(query, initNs={'xsd': XSD, 'dcterms': DCTERMS})
    g = graph_from_triples(res)
    print(f'found {len(g)} latin-1 triples in graph {graph.identifier}')
    return g


def find_latin1_preannos(graph: Graph, source_graph: Graph) -> Graph:
    query = r'''CONSTRUCT {{?selector ?pred ?obj}}
    WHERE {{
        GRAPH <{}> {{
            ?s oa:hasTarget ?target .
            ?target oa:hasSource ?source ;
            oa:hasSelector ?selector .
            ?selector ?pred ?obj
            FILTER(datatype(?obj)=xsd:string)
            FILTER(regex(?obj, "[\\x80-\\xFF]"))

        }}
        GRAPH <{}> {{
            ?source dcterms:created ?date .
            FILTER(?date > "2022-05-10T00:00:00.000000+00:00"^^xsd:dateTime)
        }}
    }}
    '''.format(graph.identifier, source_graph.identifier)
    res = graph.query(query, initNs={'xsd': XSD, 'dcterms': DCTERMS, 'oa': OA})
    g = graph_from_triples(res)
    print(f'found {len(g)} latin-1 triples in graph {graph.identifier}')
    return g

def recode_latin1_triples(g: Graph, latin1_triples: Graph, commit=False) -> None:
    '''Find and recodes latin1-encoded strings to utf-8
    If commit, also replace them in the triplestore.
    '''
    cnt = 0
    for (s, p, o) in latin1_triples:
        recoded = latin1_to_utf8(o)
        if o != recoded:
            if not commit:
                # manual sanity check
                print(o)
                print(recoded)
                print('---')
            else:
                g.add((s, p, Literal(recoded)))
                g.remove((s, p, o))
                cnt += 1
    print(f'updated {cnt} triples')

def patched_inject_prefixes(self, query, extra_bindings):
    ''' Monkeypatch for SPARQLStore prefix injection
    Parses the incoming query for prefixes,
    and ignores these when injecting additional namespaces.
    Better implementation is possibly available,
    e.g. use rdfblibs query parser to extract prefixes.
    '''
    query_prefixes = re.findall(PREFIX_PATTERN, query)

    # prefixes available in the query should be deducted from the store's nsBindings
    # prefixes that were provided through initNs should take precedence over all others
    bindings = {x for x in set(self.nsBindings.items())
                if x[0] not in query_prefixes}
    bindings |= set(extra_bindings.items())

    # remove the extra bindings from the original query
    for k in set(extra_bindings.keys()):
        if k in query_prefixes:
            replace_pattern = re.compile(
                fr'PREFIX\s+{k}:\s*<.+>', re.IGNORECASE)
            query = re.sub(replace_pattern, '', query)

    if not bindings:
        return query
    return "\n".join(
        [
            "\n".join(["PREFIX %s: <%s>" % (k, v) for k, v in bindings]),
            "",  # separate ns_bindings from query with an empty line
            query,
        ]
    )


def patched_sparqlconnector_update(self, query,
                                   default_graph: Optional[str] = None,
                                   named_graph: Optional[str] = None):
    '''Monkeypatch for SPARQLConnector's update method
    Changes Content-Type header to include utf-8 charset
    '''
    if not self.update_endpoint:
        raise SPARQLConnectorException("Query endpoint not set!")

    params = {}

    if default_graph is not None:
        params["using-graph-uri"] = default_graph

    if named_graph is not None:
        params["using-named-graph-uri"] = named_graph

    # Single difference from original method, changing Content-Type header
    headers = {
        "Accept": _response_mime_types[self.returnFormat],
        "Content-Type": "application/sparql-update; charset=utf-8",
    }

    args = dict(self.kwargs)  # other QSAs

    args.setdefault("params", {})
    args["params"].update(params)
    args.setdefault("headers", {})
    args["headers"].update(headers)

    qsa = "?" + urlencode(args["params"])
    res = urlopen(
        Request(self.update_endpoint + qsa, data=query.encode(),
                headers=args["headers"])
    )


# Apply monkeypatches
SPARQLStore._inject_prefixes = patched_inject_prefixes
SPARQLConnector.update = patched_sparqlconnector_update



from .views import UpdateView, QueryView
from rest_framework.urlpatterns import format_suffix_patterns
from django.urls import path

urlpatterns = format_suffix_patterns([
    path('test/query', QueryView.as_view()),
    path('test/update', UpdateView.as_view()),
])


from rest_framework.views import APIView, exception_handler
from rest_framework.response import Response
from rest_framework.status import *
from rest_framework.exceptions import NotFound

from rdflib import Graph, URIRef, BNode, Literal

from rdf.ns import *
from rdf.renderers import TurtleRenderer
from rdf.parsers import JSONLDParser
from rdf.utils import append_triples, graph_from_triples

HTTPSC_MAP = {
    # Both Django REST framework and HTTPSC are incomplete.
    # This mapping contains only the error codes they have in common.
    HTTP_400_BAD_REQUEST: HTTPSC.BadRequest,
    HTTP_401_UNAUTHORIZED: HTTPSC.Unauthorized,
    HTTP_402_PAYMENT_REQUIRED: HTTPSC.PaymentRequired,
    HTTP_403_FORBIDDEN: HTTPSC.Forbidden,
    HTTP_404_NOT_FOUND: HTTPSC.NotFound,
    HTTP_405_METHOD_NOT_ALLOWED: HTTPSC.MethodNotAllowed,
    HTTP_406_NOT_ACCEPTABLE: HTTPSC.NotAcceptable,
    HTTP_407_PROXY_AUTHENTICATION_REQUIRED: HTTPSC.ProxyAuthenticationRequired,
    HTTP_408_REQUEST_TIMEOUT: HTTPSC.RequestTimeout,
    HTTP_409_CONFLICT: HTTPSC.Conflict,
    HTTP_410_GONE: HTTPSC.Gone,
    HTTP_411_LENGTH_REQUIRED: HTTPSC.LengthRequired,
    HTTP_412_PRECONDITION_FAILED: HTTPSC.PreconditionFailed,
    HTTP_413_REQUEST_ENTITY_TOO_LARGE: HTTPSC.RequestEntityTooLarge,
    HTTP_414_REQUEST_URI_TOO_LONG: HTTPSC.RequestURITooLong,
    HTTP_415_UNSUPPORTED_MEDIA_TYPE: HTTPSC.UnsupportedMediaType,
    HTTP_416_REQUESTED_RANGE_NOT_SATISFIABLE: HTTPSC.RequestedRangeNotSatisfiable,
    HTTP_417_EXPECTATION_FAILED: HTTPSC.ExpectationFailed,
    HTTP_422_UNPROCESSABLE_ENTITY: HTTPSC.UnprocessableEntity,
    HTTP_423_LOCKED: HTTPSC.Locked,
    HTTP_424_FAILED_DEPENDENCY: HTTPSC.FailedDependency,
    HTTP_500_INTERNAL_SERVER_ERROR: HTTPSC.InternalServerError,
    HTTP_501_NOT_IMPLEMENTED: HTTPSC.NotImplemented,
    HTTP_502_BAD_GATEWAY: HTTPSC.BadGateway,
    HTTP_503_SERVICE_UNAVAILABLE: HTTPSC.ServiceUnavailable,
    HTTP_504_GATEWAY_TIMEOUT: HTTPSC.GatewayTimeout,
    HTTP_505_HTTP_VERSION_NOT_SUPPORTED: HTTPSC.HTTPVersionNotSupported,
    HTTP_507_INSUFFICIENT_STORAGE: HTTPSC.InsufficientStorage,
}


def graph_from_request(request):
    """ Safely obtain a graph, from the request if present, empty otherwise. """
    data = request.data
    if not isinstance(data, Graph):
        data = Graph()
    return data


def custom_exception_handler(error, context):
    """
    Returns a rest_framework Response with an rdflib Graph as .data.

    Normally, a Response would have a JSON-serializable dict or list
    as data.
    """
    response = exception_handler(error, context)
    if response is None:
        return response
    status = response.status_code
    original_data = response.data

    res = BNode()
    override_data = graph_from_triples((
        (res, RDF.type, HTTP.Response),
        (res, HTTP.statusCodeValue, Literal(status)),
    ))
    if isinstance(original_data, list):
        append_triples(override_data, (
            (res, HTTP.reasonPhrase, Literal(str(d))) for d in original_data
        ))
    else:
        override_data.add((res, HTTP.reasonPhrase, Literal(original_data['detail'])))
    status_uri = HTTPSC_MAP.get(status)
    if status_uri:
        override_data.add((res, HTTP.sc, status_uri))
    response.data = override_data
    return response


class RDFView(APIView):
    """
    Expose a given graph as an RDF-encoded API endpoint.

    Either set a static `graph` member function or override
    `get_graph` to compute the graph from the request.

    By default, this view supports Turtle as output and JSON-LD as input,
    but this can be overridden by providing custom values for
    `renderer_classes` and `parser_classes`.

    Only `JsonLdRenderer` as an alternative renderer class has been tested.
    If using this, JSON LD context may be provided by setting the
    `json_ld_context` property (generally a `dict`).
    """
    renderer_classes = (TurtleRenderer,)
    parser_classes = (JSONLDParser,)

    def get(self, request, format=None, **kwargs):
        return Response(self.get_graph(request, **kwargs))

    def get_graph(self, request, **kwargs):
        return self.graph()

    def get_exception_handler(self):
        return custom_exception_handler


class RDFResourceView(RDFView):
    """ API endpoint for fetching individual subjects. """

    def get(self, request, format=None, **kwargs):
        data = self.get_graph(request, **kwargs)
        if len(data) == 0:
            raise NotFound()
        return Response(data)

    def get_graph(self, request, **kwargs):
        identifier = URIRef(self.get_resource_uri(request, **kwargs))
        pattern = (identifier, None, None)
        result = graph_from_triples(self.graph().triples(pattern))
        # TODO: also include related nodes
        return result

    def get_resource_uri(self, request, **kwargs):
        return request.build_absolute_uri(request.path)



from rdflib import URIRef

MIGRATION_OVERRIDE_MESSAGE = 'Override this method to return an rdflib.Graph.'


def on_add(uri):
    """
    Decorator for RDFMigration class methods.

    Use this to mark a method as handling the introduction of a
    particular subject URI. Each method can be decorated only once in
    this way, although it is possible to mark a method as handling
    both an addition and a removal. Conversely, each URI can be
    handled by only one method, although both its addition and its
    removal can be handled.
    """
    def decorator(method):
        method.__handles_addition__ = URIRef(uri)
        return method
    return decorator


def on_remove(uri):
    """
    Decorator for RDFMigration class methods.

    Use this to mark a method as handling the removal of a particular
    subject URI. Each method can be decorated only once in this way,
    although it is possible to mark a method as handling both an
    addition and a removal. Conversely, each URI can be handled by
    only one method, although both its addition and its removal can
    be handled.
    """
    def decorator(method):
        method.__handles_removal__ = URIRef(uri)
        return method
    return decorator


def on_present(uri):
    """
    Decorator for RDFMigration class methods.

    Use this to mark a method as handling the presence of a
    particular predicate URI in the actual graph. Each method can
    be decorated only once in this way.
    """

    def decorator(method):
        method.__handles_presence__ = URIRef(uri)
        return method
    return decorator


class MetaRDFMigration(type):
    """
    Metaclass that ensures the presence of .add_handlers et al.

    This should be considered an implementation detail of RDFMigration.
    """

    def __new__(cls, name, bases, namespace, **kwargs):
        augmented = super().__new__(cls, name, bases, namespace, **kwargs)
        add_handlers = {}
        del_handlers = {}
        pres_handlers = {}
        for attribute_name, value in augmented.__dict__.items():
            add_uri = getattr(value, '__handles_addition__', None)
            del_uri = getattr(value, '__handles_removal__', None)
            pres_uri = getattr(value, '__handles_presence__', None)
            if add_uri:
                add_handlers[add_uri] = attribute_name
            if del_uri:
                del_handlers[del_uri] = attribute_name
            if pres_uri:
                pres_handlers[pres_uri] = attribute_name
        augmented.add_handlers = add_handlers
        augmented.remove_handlers = del_handlers
        augmented.presence_handlers = pres_handlers

        return augmented


class RDFMigration(metaclass=MetaRDFMigration):
    """
    Base class for RDF migrations.

    At a minimum, every derived class should override the .actual and
    .desired member functions, both of which must return an
    rdflib.Graph. If you are assigning free functions to these
    members, for example imported from another module, make sure to
    wrap them with staticmethod().

    By default, .actual() will be updated to match .desired()
    exactly. If you need to do extra work, define additional methods
    that are decorated with @on_add or @on_remove (documented above).
    These methods should take three arguments, `self`, `actual` and
    `conjunctive`, where `actual` is the result of `self.actual()`
    and `conjunctive` is the full conjunctive graph from the Django
    store. The return value is ignored; migration methods should take
    effect by directly modifying either `actual` or `conjunctive`.

    Migrations are executed in the following order:

      1. Add new triples from the desired graph to the actual graph.
      2. Execute addition handler methods (relative order undefined).
      3. Execute removal handler methods (relative order undefined).
      4. Remove triples from the actual that disappeared in the desired.

    A migration class is instantiated before application, so in
    principle, you can make a migration stateful. However, keep in
    mind that the relative order of addition handlers and removal
    handlers is undefined; for this reason, it is safer to keep the
    migration stateless if possible.

    If you define an __init__ method, give it a (self, *args, **kwargs)
    signature in order to make it forward-compatible. Currently, no
    arguments are passed, but this might change in the future.
    """

    def actual(self):
        raise NotImplementedError(MIGRATION_OVERRIDE_MESSAGE)

    def desired(self):
        raise NotImplementedError(MIGRATION_OVERRIDE_MESSAGE)



import re

from django.http.response import HttpResponseBase
from pyparsing import ParseException
from rdf.ns import HTTP, HTTPSC, RDF
from rdf.renderers import TurtleRenderer
from rdf.utils import graph_from_triples
from rdf.views import custom_exception_handler as turtle_exception_handler
from rdflib import BNode, Literal
from rdflib.plugins.sparql.parser import parseUpdate
from requests.exceptions import HTTPError
from urllib.error import HTTPError as urllibHTTPError
from rest_framework.exceptions import APIException, NotAcceptable, ParseError
from rest_framework.response import Response
from rest_framework.views import APIView

from .constants import (BLANK_NODE_PATTERN, SPARQL_NS, UPDATE_NOT_SUPPORTED,
                        UPDATE_NOT_SUPPORTED_PATTERN)
from .exceptions import (BlankNodeError, NoParamError, ParseSPARQLError,
                         UnsupportedUpdateError)
from .negotiation import SPARQLContentNegotiator



class SPARQLUpdateAPIView(APIView):
    '''
    Parent class for an SPARL update request.
    '''
    
    renderer_classes = (TurtleRenderer,)

    def get_exception_handler(self):
        return turtle_exception_handler

    def check_supported(self, updatestring):
        # check the entire string for unsupported keywords
        if re.match(UPDATE_NOT_SUPPORTED_PATTERN, updatestring):
            # do a dry-run parse of the updatestring
            parse_request = parseUpdate(updatestring).request
            # check if the parse contains any unsupported operations
            if any(part.name in UPDATE_NOT_SUPPORTED for part in parse_request):
                raise UnsupportedUpdateError(
                    'Update operation is not supported.'
                )

        # Do a quick check for blank nodes
        if re.search(BLANK_NODE_PATTERN, updatestring):
            parse_request = parseUpdate(updatestring).request
            for part in parse_request:
                if any(
                        isinstance(term, BNode)
                        for triple in part.quads.triples
                        for term in triple
                ):
                    raise BlankNodeError()

        return

    def execute_update(self, updatestring):
        graph = self.graph()

        try:
            self.check_supported(updatestring)
            graph.update(updatestring)
        except (ParseException, ParseError, ValueError) as p_e:
            # Raised when SPARQL syntax is not valid, or parsing fails
            graph.rollback()
            raise ParseSPARQLError(p_e)
        except HTTPError as h_e:
            graph.rollback()
            if 400 <= h_e.response.status_code < 500:
                raise ParseSPARQLError(h_e)
            else:
                raise APIException(h_e)
        except urllibHTTPError as u_h_e:
            graph.rollback()
            if 400 <= u_h_e.code < 500:
                raise ParseSPARQLError(u_h_e)
            else:
                raise APIException(u_h_e)
        except Exception as e:
            graph.rollback()
            raise APIException(e)

    def post(self, request, **kwargs):
        """ Accepts POST request with SPARQL-Query in body parameter 'query'
            Renders text/turtle
        """
        sparql_string = request.data.get("update")
        if not sparql_string:
            # POST must contain an update
            raise NoParamError()
        blank = BNode()
        status = 200
        response = graph_from_triples(
            (
                (blank, RDF.type, HTTP.Response),
                (blank, HTTP.statusCodeValue, Literal(status)),
                (blank, HTTP.reasonPhrase, Literal("Updated successfully")),
                (blank, HTTP.sc, HTTPSC.OK),
            )
        )
        self.execute_update(sparql_string)

        return Response(response)

    def graph(self):
        raise NotImplementedError


class SPARQLQueryAPIView(APIView):
    '''
    Parent class for a SPARQL query request.
    '''
    
    renderer_classes = SPARQLContentNegotiator.rdf_renderers + \
        SPARQLContentNegotiator.results_renderers
    content_negotiation_class = SPARQLContentNegotiator

    def get_exception_handler(self):
        ''' Errors are returned as turtle, even when the original querytype
        does not satisfy could not be rendered in this format. A hard
        overwrite of renderer is necessary. Ideally, this should render errors
        in a querytype-compatibleway. '''
        self.request.accepted_renderer = TurtleRenderer()
        self.request.accepted_media_type = TurtleRenderer.media_type
        return turtle_exception_handler

    def execute_query(self, querystring):
        """ Attempt to query a graph with a SPARQL-Query string
            Sets query type on succes
        """
        graph = self.graph()
        try:
            if not querystring:
                query_results = graph
                query_type = "EMPTY"
            else:
                # See SPARQLUpdateAPIView.execute_update
                query_results = graph.query(querystring)
                query_type = query_results.type
            self.request.data["query_type"] = query_type

            # re-perform content negotiation to determine if
            # querytype satisfies accept header
            neg = self.perform_content_negotiation(self.request)
            self.request.accepted_renderer, self.request.accepted_media_type = neg
            return query_results

        except (ParseException, ValueError) as p_e:
            # Raised when SPARQL syntax is not valid, or parsing fails
            graph.rollback()
            raise ParseSPARQLError(p_e)
        except HTTPError as h_e:
            graph.rollback()
            if 400 <= h_e.response.status_code < 500:
                raise ParseSPARQLError(h_e)
            else:
                raise APIException(h_e)
        except urllibHTTPError as u_h_e:
            graph.rollback()
            if 400 <= u_h_e.code < 500:
                raise ParseSPARQLError(u_h_e)
            else:
                raise APIException(u_h_e)
        except NotAcceptable:
            graph.rollback()
            raise
        except Exception as n_e:
            graph.rollback()
            raise APIException(n_e)

    def get(self, request, **kwargs):
        """ Accepts GET request, optional SPARQL-Query
            in query parameter 'query'.
            Without 'query' parameter, returns the entire graph as text/turtle.
            Renders application/json or text/turtle based
            on query type and header 'Accept.
        """
        sparql_string = request.query_params.get("query")
        if request.data:
            raise ParseSPARQLError(
                "GET request should provide query in parameter, not request body.")
        query_results = self.execute_query(sparql_string)

        return Response(query_results)

    def post(self, request, **kwargs):
        """ Accepts POST request with SPARQL-Query in body parameter 'query'.
            Renders application/json or text/turtle based
            on query type and header 'Accept'.
        """
        sparql_string = request.data.get("query")
        if not sparql_string:
            raise NoParamError()
        # request.data is immutable for POST requests
        request.data._mutable = True
        query_results = self.execute_query(sparql_string)

        return Response(query_results)

    def graph(self):
        raise NotImplementedError



from importlib import import_module
from datetime import datetime, date

from django.conf import settings
from pytest import fixture
from rdf.utils import graph_from_triples, prune_triples
from rdflib.plugins.stores.sparqlstore import SPARQLUpdateStore
from rdflib import ConjunctiveGraph, Graph, Literal, URIRef

from .ns import *

CREATION_DATE = Literal(datetime.now())

ITEM = Namespace('http://localhost:8000/item/')
STAFF = Namespace('http://localhost:8000/staff#')
ONTO = Namespace('http://localhost:8000/ontology#')
SOURCE = Namespace('http://localhost:8000/source/')

ITEMS = (
    ( ITEM['1'], RDF.type,              OA.TextQuoteSelector               ),
    ( ITEM['1'], OA.prefix,             Literal('this is the start of ')   ),
    ( ITEM['1'], OA.exact,              Literal('the exact selection')     ),
    ( ITEM['1'], OA.suffix,             Literal(' and this is the end')    ),
    ( ITEM['1'], DCTERMS.creator,       STAFF.tester                       ),
    ( ITEM['1'], DCTERMS.created,       CREATION_DATE                      ),

    # Items 2 and 3 dropped out for historical reasons.

    ( ITEM['4'], RDF.type,              OA.TextPositionSelector            ),
    ( ITEM['4'], OA.start,              Literal(22)                        ),
    ( ITEM['4'], OA.end,                Literal(41)                        ),
    ( ITEM['4'], DCTERMS.creator,       STAFF.tester                       ),
    ( ITEM['4'], DCTERMS.created,       CREATION_DATE                      ),

    ( ITEM['5'], RDF.type,              OA.SpecificResource                ),
    ( ITEM['5'], OA.hasSource,          SOURCE['1']                        ),
    ( ITEM['5'], OA.hasSelector,        ITEM['4']                          ),
    ( ITEM['5'], OA.hasSelector,        ITEM['1']                          ),
    ( ITEM['5'], DCTERMS.creator,       STAFF.tester                       ),
    ( ITEM['5'], DCTERMS.created,       CREATION_DATE                      ),

    ( ITEM['6'], RDF.type,              ONTO.reader                        ),
    ( ITEM['6'], SKOS.prefLabel,        Literal('Margaret Blessington')    ),
    ( ITEM['6'], ONTO.is_identified_by, Literal('Margaret Gardiner Blessington') ),
    ( ITEM['6'], ONTO.is_identified_by, Literal('Marguerite Gardiner, countess of Blessington') ),
    ( ITEM['6'], ONTO.has_gender,       Literal('female')                  ),
    ( ITEM['6'], ONTO.has_occupation,   Literal('novelist')                ),
    ( ITEM['6'], ONTO.has_occupation,   Literal('writer')                  ),
    ( ITEM['6'], ONTO.has_nationality,  Literal('Irish')                   ),
    ( ITEM['6'], CIDOC.was_born,        Literal(date(1789, 9, 1))          ),
    ( ITEM['6'], CIDOC.died,            Literal(date(1849, 6, 4))          ),
    ( ITEM['6'], OWL.sameAs,            URIRef('https://en.wikipedia.org/wiki/Marguerite_Gardiner,_Countess_of_Blessington') ),
    ( ITEM['6'], DCTERMS.type,          Literal('example data')            ),
    ( ITEM['6'], DCTERMS.creator,       STAFF.tester                       ),
    ( ITEM['6'], DCTERMS.created,       CREATION_DATE                      ),

    ( ITEM['7'], RDF.type,              OA.Annotation                      ),
    ( ITEM['7'], OA.hasBody,            ONTO.reader                        ),
    ( ITEM['7'], OA.hasBody,            ITEM['6']                          ),
    ( ITEM['7'], OA.hasTarget,          ITEM['5']                          ),
    ( ITEM['7'], OA.motivatedBy,        OA.tagging                         ),
    ( ITEM['7'], OA.motivatedBy,        OA.identifying                     ),
    ( ITEM['7'], DCTERMS.creator,       STAFF.tester                       ),
    ( ITEM['7'], DCTERMS.created,       CREATION_DATE                      ),
)


MAGIC_NODE = URIRef('http://hogwarts.edu/')

TRIPLES = (
    (RDF.type, RDF.type, RDF.Property),
    (RDF.type, RDFS.range, RDFS.Class),
    (RDF.type, RDFS.label, Literal('type')),
    (RDFS.range, RDFS.domain, RDF.Property),
    (RDFS.domain, RDFS.domain, RDF.Property),
    # following triples added for testing traversal
    # note that these form a ring together with the triples above!
    (RDF.Property, RDF.type, RDFS.Class),
    (RDF.Property, RDFS.subClassOf, RDFS.Resource),
    (RDFS.Resource, RDFS.isDefinedBy, MAGIC_NODE),
    (MAGIC_NODE, RDF.type, RDF.Statement),
    (MAGIC_NODE, RDF.object, RDF.type),
)


@fixture
def triples():
    return TRIPLES


@fixture
def empty_graph():
    return Graph()


@fixture
def items():
    return ITEMS


@fixture
def filled_graph(triples):
    return graph_from_triples(triples)


@fixture
def filled_conjunctive_graph(items):
    return graph_from_triples(items, ConjunctiveGraph)


@fixture
def app_with_rdf_migrations():
    from .test_apps import with_migrations
    yield with_migrations.__name__
    # Make sure the graph is empty again after use
    graph = import_module('.graph', with_migrations.__name__)
    g = graph.graph()
    prune_triples(g, g)
    assert len(g) == 0


@fixture
def app_without_rdf_migrations():
    from .test_apps import without_migrations
    return without_migrations.__name__


@fixture
def prefixed_query():
    return '''
    PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    PREFIX schema: <http://www.schema.org/>
    SELECT ?s ?p ?o WHERE { ?s ?p ?o }
    '''


HAS_TRIPLES = '''
ASK {
    GRAPH ?g {
        ?s ?p ?o
    }
}
'''

@fixture
def sparqlstore(settings):
    store = settings.RDFLIB_STORE
    store.update('CLEAR ALL')
    assert not store.query(HAS_TRIPLES)
    yield store
    store.update('CLEAR ALL')




"""
Algorithms and manage.py command for migrating RDF data.

Django applications that wish to support this command should include
an rdf_migrations module which should define a Migration class that
derives from rdf.migrations.RDFMigration.
"""

import logging
from importlib import import_module

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from rdf.utils import append_triples, get_conjunctive_graph, prune_triples

logger = logging.getLogger('rdf')


class Command(BaseCommand):
    help = 'Performs RDF migrations for the specified app(s).'

    def add_arguments(self, parser):
        parser.add_argument('app_label', nargs='*')

    def handle(self, *args, **options):
        app_labels = options['app_label']
        if len(app_labels) == 0:
            app_labels = settings.INSTALLED_APPS
        else:
            for app in app_labels:
                if app not in settings.INSTALLED_APPS:
                    raise CommandError('{} is not an installed app'.format(app))
        for app in app_labels:
            self.migrate_package(app)

    def migrate_package(self, pkg_name):
        """ Determine whether `pkg` has RDF migrations. If so, apply them. """
        try:
            migrations = import_module('.rdf_migrations', pkg_name)
            if migrations:
                logger.info(
                    'Applying RDF migrations for {}...'.format(pkg_name))
            self.migrate_graph(migrations.Migration())
            applied_msg = 'Applied RDF migrations for {}.'.format(pkg_name)
            self.stdout.write(applied_msg)
            logger.info(applied_msg)
        except ImportError:
            # Nothing to do, this package has no RDF migrations.
            pass
        except AttributeError:
            # Likewise.
            pass

    def migrate_graph(self, migration):
        """ Update the `actual` graph to match `desired`. """
        actual = migration.actual()
        desired = migration.desired()
        additions = desired - actual
        deletions = actual - desired
        logger.info('{} additions, {} deletions.'.format(
            len(additions), len(deletions)))
        predicates_present = set(actual.predicates())
        subjects_added = set(additions.subjects())
        subjects_deleted = set(deletions.subjects())
        conjunctive = get_conjunctive_graph()
        # Do the additions first in case we need to update referencing triples.
        append_triples(actual, additions)
        adders = (migration.add_handlers.get(s) for s in subjects_added)
        for handler_name in filter(None, adders):
            getattr(migration, handler_name)(actual, conjunctive)
            logger.info('Applied addition handler "{}".'.format(handler_name))
        deleters = (migration.remove_handlers.get(s) for s in subjects_deleted)
        for handler_name in filter(None, deleters):
            getattr(migration, handler_name)(actual, conjunctive)
            logger.info('Applied deletion handler "{}".'.format(handler_name))
        prune_triples(actual, deletions)
        # Handle predicate presence
        presence_handlers = (migration.presence_handlers.get(s)
                             for s in predicates_present)
        for handler_name in filter(None, presence_handlers):
            getattr(migration, handler_name)(actual, conjunctive)
            logger.info('Applied presence handler "{}".'.format(handler_name))



from rdflib.graph import Graph
from django.conf import settings
from sparql.test_app.constants import *

def sources_graph():
    return Graph(settings.RDFLIB_STORE, SOURCES_NS)

def nlp_graph():
    return Graph(settings.RDFLIB_STORE, nlp)